import torch
import torchdiffeq
import time
import scipy as sp
import numpy as np
import sys
import json
import gc

import traceback


# Ivp Solver with GPU
class IvpSolverPytorch:
    def __init__(self):
        pass

    def solve(
        self,
        t: torch.Tensor,
        psi0: torch.Tensor,
        rtol: float = 1e-9,
        atol: float = 1e-9,
        method: str = "dopri5",
    ):
        res = torchdiffeq.odeint(self.func, psi0, t, rtol=rtol, atol=atol, method=method)
        return res

    def func(self, t, y: torch.Tensor):
        rhs = -y * torch.sin(t)
        # rhs = torch.empty_like(y)
        # rhs[0::2] = y[1::2]
        # rhs[1::2] = ((1.0 - y[0::2] ** 2) * y[1::2] - y[0::2]) / 1e-6
        return rhs


# Ivp Solver with odeint
class IvpSolverScipyOdeint:
    def __init__(self):
        pass

    def solve(
        self,
        t: np.array,
        y0: np.array,
        rtol: float = 1e-9,
        atol: float = 1e-9,
    ):
        res = sp.integrate.odeint(self.func, y0, t, atol=atol, rtol=rtol)

        return res

    def func(self, y: np.array, t: float):
        rhs = -y * np.sin(t)
        # rhs = np.empty_like(y)
        # rhs[0::2] = y[1::2]
        # rhs[1::2] = ((1.0 - y[0::2] ** 2) * y[1::2] - y[0::2]) / 1e-6
        return rhs


# Ivp Solver with CPU
class IvpSolverScipy:
    def __init__(self):
        pass

    def solve(
        self,
        t: np.array,
        y0: np.array,
        rtol: float = 1e-9,
        atol: float = 1e-9,
        method: str = "RK45",
    ):
        res = sp.integrate.solve_ivp(self.func, [t[0], t[-1]], y0, t_eval=t, rtol=rtol, atol=atol, method=method)
        return res

    def func(self, t: float, y: np.array):
        rhs = -y * np.sin(t)
        # rhs = np.empty_like(y)
        # rhs[0::2] = y[1::2]
        # rhs[1::2] = ((1.0 - y[0::2] ** 2) * y[1::2] - y[0::2]) / 1e-6
        return rhs


# --- Benchmark結果を保持するクラス ---
class BenchmarkResult:
    def __init__(self, N):
        self.N = N

        self.sp_odeint_LSODA = []
        self.sp_solveivp_RK23 = []
        self.sp_solveivp_RK45 = []
        self.sp_solveivp_DOP853 = []
        self.sp_solveivp_Radau = []
        self.sp_solveivp_BDF = []
        self.sp_solveivp_LSODA = []
        self.tc_odeint_dopri8 = []
        self.tc_odeint_dopri5 = []
        self.tc_odeint_bosh3 = []
        self.tc_odeint_fehlberg2 = []
        self.tc_odeint_adaptive_heun = []

        self.successful_runs = 0
        self.failed_runs = 0

    def add_result(
        self,
        r_sp_odeint_LSODA,
        r_sp_solveivp_RK23,
        r_sp_solveivp_RK45,
        r_sp_solveivp_DOP853,
        r_sp_solveivp_Radau,
        r_sp_solveivp_BDF,
        r_sp_solveivp_LSODA,
        r_tc_odeint_dopri8,
        r_tc_odeint_dopri5,
        r_tc_odeint_bosh3,
        r_tc_odeint_fehlberg2,
        r_tc_odeint_adaptive_heun,
    ):
        self.sp_odeint_LSODA.append(r_sp_odeint_LSODA)
        self.sp_solveivp_RK23.append(r_sp_solveivp_RK23)
        self.sp_solveivp_RK45.append(r_sp_solveivp_RK45)
        self.sp_solveivp_DOP853.append(r_sp_solveivp_DOP853)
        self.sp_solveivp_Radau.append(r_sp_solveivp_Radau)
        self.sp_solveivp_BDF.append(r_sp_solveivp_BDF)
        self.sp_solveivp_LSODA.append(r_sp_solveivp_LSODA)
        self.tc_odeint_dopri8.append(r_tc_odeint_dopri8)
        self.tc_odeint_dopri5.append(r_tc_odeint_dopri5)
        self.tc_odeint_bosh3.append(r_tc_odeint_bosh3)
        self.tc_odeint_fehlberg2.append(r_tc_odeint_fehlberg2)
        self.tc_odeint_adaptive_heun.append(r_tc_odeint_adaptive_heun)
        self.successful_runs += 1

    def add_failure(self):
        self.failed_runs += 1

    def compute_stats(self):
        stats = {}
        for key, values in {
            "ScipyOdeintLSODA": self.sp_odeint_LSODA,
            "ScipySolveivpRK23": self.sp_solveivp_RK23,
            "ScipySolveivpRK45": self.sp_solveivp_RK45,
            "ScipySolveivpDOP853": self.sp_solveivp_DOP853,
            "ScipySolveivpRadau": self.sp_solveivp_Radau,
            "ScipySolveivpBDF": self.sp_solveivp_BDF,
            "ScipySolveivpLSODA": self.sp_solveivp_LSODA,
            "PytorchOdeintDopri8": self.tc_odeint_dopri8,
            "PytorchOdeintDopri5": self.tc_odeint_dopri5,
            "PytorchOdeintBosh3": self.tc_odeint_bosh3,
            "PytorchOdeintFehlberg2": self.tc_odeint_fehlberg2,
            "PytorchOdeintAdaptive_heun": self.tc_odeint_adaptive_heun,
        }.items():
            if values:
                # 全要素が -1.0 の場合は処理しない
                if np.all(values == -1.0):
                    mean = -1.0
                    std = -1.0
                else:
                    # -1.0が中途半端に含まれる場合、それを除外して計算
                    valid_values = values[values != -1.0]
                    valid_values = []
                    for v in values:
                        if v > -0.5:
                            valid_values.append(v)

                    if valid_values:
                        mean = np.mean(valid_values)
                        std = np.std(valid_values)
                    else:
                        mean = -1.0
                        std = -1.0

            else:
                mean, std = -1.0, -1.0
            stats[key] = {
                "times": values,
                "mean": mean,
                "std": std,
            }
        stats["successful_runs"] = self.successful_runs
        stats["failed_runs"] = self.failed_runs
        return stats

    def to_dict(self):
        return {"N": self.N, "stats": self.compute_stats()}


# Main関数
if __name__ == "__main__":
    # N依存性の調査対象
    N1 = np.logspace(2, 16, num=15, base=2)  # 2^2 ~ 2^22 を対数スケールで41分割
    N2 = np.logspace(1, 5, num=5, base=10)  # 10^1 ~ 10^6 を対数スケールで21分割
    N_values = np.unique(np.round(np.concatenate([N1, N2])).astype(int)).tolist()

    limit_LSODA_N = 10000  # LSODAの場合、メモリがOut of rangeになるため、制限をかける
    calc_time_threshold = 30.0  # calc_time_threshold 秒以上かかった場合、以降の計算を行わない

    benchmark_sets = 10  # 各Nでのセット数

    # 時間範囲の設定
    t_arr = np.linspace(0, 2.0, 11, endpoint=True)
    t_gpu = torch.tensor(t_arr, device="cuda", dtype=torch.float64)

    # パラメータの設定
    tols = [1e-4, 1e-6, 1e-9]
    for tol in tols:
        rtol = tol
        atol = tol

        # 結果出力先ファイル（各N毎に追記）
        result_filename = f"results_tol_{tol}.jsonl"

        is_keep_calc = {
            "odeintLSODA": True,
            "RK23": True,
            "RK45": True,
            "DOP853": True,
            "Radau": True,
            "BDF": True,
            "LSODA": True,
            "dopri8": True,
            "dopri5": True,
            "bosh3": True,
            "fehlberg2": True,
            "adaptive_heun": True,
        }

        # ループ開始
        for N in N_values:
            print(f"\n=== ベンチマーク開始: N = {N} ===")
            bm_result = BenchmarkResult(N)
            for i in range(benchmark_sets):
                print(f"--- セット {i+1:02d} / {benchmark_sets} ---")
                try:
                    # 初期状態の設定 (各ルーチンで同一のy0を用いる)
                    y0 = np.full(N, 2.0, dtype=np.float64)
                    # y0[0::2] = 2.0
                    # y0[1::2] = -0.66
                    y0_gpu = torch.full((N,), 2.0, device="cuda", dtype=torch.float64)
                    # y0_gpu[0::2] = 2.0
                    # y0_gpu[1::2] = -0.66

                    time_results = []

                    # ----- scipy.odeint -----
                    method_list = ["odeintLSODA"]
                    for method in method_list:
                        if method == "odeintLSODA" and N > limit_LSODA_N:
                            is_keep_calc[method] = False

                        if is_keep_calc[method]:
                            solver = IvpSolverScipyOdeint()
                            t_start = time.time()
                            solution = solver.solve(t_arr, y0, rtol=rtol, atol=atol)
                            elapsed_time = time.time() - t_start
                            time_results.append(elapsed_time)
                            print(f"  Elapsed time Scipy Odeint {method:15s} : {elapsed_time:10.6f} seconds")
                            if elapsed_time > calc_time_threshold:
                                is_keep_calc[method] = False
                            del solver
                            del solution
                            gc.collect()
                        else:
                            time_results.append(-1.0)

                    # ----- scipy solve_ivp -----
                    method_list = ["RK23", "RK45", "DOP853", "Radau", "BDF", "LSODA"]
                    for method in method_list:
                        if method == "LSODA" and N > limit_LSODA_N:
                            is_keep_calc[method] = False

                        if is_keep_calc[method]:
                            solver = IvpSolverScipy()
                            t_start = time.time()
                            solution = solver.solve(t_arr, y0, rtol=rtol, atol=atol, method=method)
                            elapsed_time = time.time() - t_start
                            time_results.append(elapsed_time)
                            print(f"  Elapsed time Scipy solve_ivp {method:12s} : {elapsed_time:10.6f} seconds")
                            if elapsed_time > calc_time_threshold:
                                is_keep_calc[method] = False
                            del solver
                            del solution
                            gc.collect()
                        else:
                            time_results.append(-1.0)

                    #  print(f"  Elapsed time Ivp CPU: {ivp_cpu_time:10.6f} seconds")

                    # ----- Ivp GPU -----
                    method_list = ["dopri8", "dopri5", "bosh3", "fehlberg2", "adaptive_heun"]
                    for method in method_list:
                        if is_keep_calc[method]:
                            solver = IvpSolverPytorch()
                            t_start = time.time()
                            solution = solver.solve(t_gpu, y0_gpu, rtol=rtol, atol=atol, method=method)
                            elapsed_time = time.time() - t_start
                            time_results.append(elapsed_time)
                            print(f"  Elapsed time Pytorch odeint {method:13s} : {elapsed_time:10.6f} seconds")
                            if elapsed_time > calc_time_threshold:
                                is_keep_calc[method] = False
                            del solver
                            del solution
                            gc.collect()
                        else:
                            time_results.append(-1.0)

                    # 結果をBenchmarkResultに追加
                    bm_result.add_result(
                        time_results[0],
                        time_results[1],
                        time_results[2],
                        time_results[3],
                        time_results[4],
                        time_results[5],
                        time_results[6],
                        time_results[7],
                        time_results[8],
                        time_results[9],
                        time_results[10],
                        time_results[11],
                    )

                except Exception as e:
                    bm_result.add_failure()
                    print(f"  セット {i+1:02d} にて例外発生:")
                    traceback.print_exc()
                    # 失敗時はそのセットの結果は記録せず，ループ継続

            # 各Nの50セット終了後，結果をファイルに追記保存
            result_dict = bm_result.to_dict()
            with open(result_filename, "a") as f:
                json.dump(result_dict, f)
                f.write("\n")
            print(f"=== N = {N} の結果を {result_filename} に保存 ===")

    sys.exit()
